@@ -32,6 +32,8 @@
 
 #include <gdbus.h>
 
+#include <ell/ell.h>
+
 #include "ofono.h"
 
 #define SHUTDOWN_GRACE_SECONDS 10
@@ -137,10 +139,19 @@ static gboolean option_version = FALSE;
 static gboolean parse_debug(const char *key, const char *value,
 					gpointer user_data, GError **error)
 {
-	if (value)
-		option_debug = g_strdup(value);
-	else
+	if (value) {
+		if (option_debug) {
+			char *prev = option_debug;
+
+			option_debug = g_strconcat(prev, ",", value, NULL);
+			g_free(prev);
+		} else {
+			option_debug = g_strdup(value);
+		}
+	} else {
+		g_free(option_debug);
 		option_debug = g_strdup("*");
+	}
 
 	return TRUE;
 }
@@ -161,6 +172,30 @@ static GOptionEntry options[] = {
 	{ NULL },
 };
 
+struct ell_event_source {
+	GSource source;
+	GPollFD pollfd;
+};
+
+static gboolean event_prepare(GSource *source, gint *timeout)
+{
+	int r = l_main_prepare();
+	*timeout = r;
+
+	return FALSE;
+}
+
+static gboolean event_check(GSource *source)
+{
+	l_main_iterate(0);
+	return FALSE;
+}
+
+static GSourceFuncs event_funcs = {
+	.prepare = event_prepare,
+	.check = event_check,
+};
+
 int main(int argc, char **argv)
 {
 	GOptionContext *context;
@@ -168,11 +203,7 @@ int main(int argc, char **argv)
 	DBusConnection *conn;
 	DBusError error;
 	guint signal;
-
-#ifdef NEED_THREADS
-	if (g_thread_supported() == FALSE)
-		g_thread_init(NULL);
-#endif
+	struct ell_event_source *source;
 
 	context = g_option_context_new(NULL);
 	g_option_context_add_main_entries(context, options, NULL);
@@ -204,12 +235,19 @@ int main(int argc, char **argv)
 
 	event_loop = g_main_loop_new(NULL, FALSE);
 
-#ifdef NEED_THREADS
-	if (dbus_threads_init_default() == FALSE) {
-		fprintf(stderr, "Can't init usage of threads\n");
-		exit(1);
-	}
-#endif
+	l_log_set_stderr();
+	l_debug_enable("*");
+	l_main_init();
+
+	source = (struct ell_event_source *) g_source_new(&event_funcs,
+					sizeof(struct ell_event_source));
+
+	source->pollfd.fd = l_main_get_epoll_fd();
+	source->pollfd.events = G_IO_IN | G_IO_HUP | G_IO_ERR;
+
+	g_source_add_poll((GSource *)source, &source->pollfd);
+	g_source_attach((GSource *) source,
+					g_main_loop_get_context(event_loop));
 
 	signal = setup_signalfd();
 
@@ -239,8 +277,6 @@ int main(int argc, char **argv)
 
 	__ofono_manager_init();
 
-	__ofono_handsfree_audio_manager_init();
-
 	__ofono_plugin_init(option_plugin, option_noplugin);
 
 	g_free(option_plugin);
@@ -250,8 +286,6 @@ int main(int argc, char **argv)
 
 	__ofono_plugin_cleanup();
 
-	__ofono_handsfree_audio_manager_cleanup();
-
 	__ofono_manager_cleanup();
 
 	__ofono_modemwatch_cleanup();
@@ -262,9 +296,14 @@ int main(int argc, char **argv)
 cleanup:
 	g_source_remove(signal);
 
+	g_source_destroy((GSource *) source);
+	l_main_exit();
+
 	g_main_loop_unref(event_loop);
 
 	__ofono_log_cleanup();
 
+	g_free(option_debug);
+
 	return 0;
 }
