@@ -23,10 +23,10 @@
 #include <config.h>
 #endif
 
-#define _GNU_SOURCE
 #include <string.h>
 #include <stdio.h>
 #include <stdint.h>
+#include <stdbool.h>
 
 #include <glib.h>
 #include <gdbus.h>
@@ -36,6 +36,8 @@
 #include <errno.h>
 #include <unistd.h>
 
+#include <ell/ell.h>
+
 #include "ofono.h"
 
 #include "common.h"
@@ -46,18 +48,33 @@
 #include "simfs.h"
 #include "stkutil.h"
 
-#define SIM_FLAG_READING_SPN	0x1
+/*
+ * A new session object will be created if a USim/ISim applications are
+ * found during app discovery. Any concurrent file/logical access to
+ * these applications will share the same session ID.
+ */
+enum session_state {
+	SESSION_STATE_INACTIVE,
+	SESSION_STATE_OPENING,
+	SESSION_STATE_CLOSING,
+	SESSION_STATE_OPEN
+};
+
+struct ofono_sim_aid_session {
+	struct sim_app_record *record;
+	int session_id;
+	struct ofono_sim *sim;
+	struct ofono_watchlist *watches;
+	enum session_state state;
+};
 
 struct ofono_sim {
-	int flags;
-
 	/* Contents of the SIM file system, in rough initialization order */
 	char *iccid;
 
 	char **language_prefs;
 	unsigned char *efli;
 	unsigned char efli_length;
-	gboolean language_prefs_update;
 
 	enum ofono_sim_password_type pin_type;
 	gboolean locked_pins[OFONO_SIM_PASSWORD_SIM_PUK]; /* Number of PINs */
@@ -74,8 +91,6 @@ struct ofono_sim {
 	unsigned char efest_length;
 	unsigned char *efsst;
 	unsigned char efsst_length;
-	gboolean fixed_dialing;
-	gboolean barred_dialing;
 
 	char *imsi;
 	char mcc[OFONO_MAX_MCC_LENGTH + 1];
@@ -87,7 +102,6 @@ struct ofono_sim {
 	unsigned char efmsisdn_records;
 
 	GSList *service_numbers;
-	gboolean sdn_ready;
 
 	unsigned char *efimg;
 	unsigned short efimg_length;
@@ -103,8 +117,10 @@ struct ofono_sim {
 	unsigned int cphs_spn_short_watch;
 
 	struct sim_fs *simfs;
+	struct sim_fs *simfs_isim;
 	struct ofono_sim_context *context;
 	struct ofono_sim_context *early_context;
+	struct ofono_sim_context *isim_context;
 
 	unsigned char *iidf_image;
 	unsigned int *iidf_watch_ids;
@@ -114,6 +130,26 @@ struct ofono_sim {
 	void *driver_data;
 	struct ofono_atom *atom;
 	unsigned int hfp_watch;
+
+	unsigned int card_slot_count;
+	unsigned int active_card_slot;
+	unsigned int pending_active_card_slot;
+
+	GSList *aid_sessions;
+	GSList *aid_list;
+	char *impi;
+	bool reading_spn : 1;
+	bool language_prefs_update : 1;
+	bool fixed_dialing : 1;
+	bool barred_dialing : 1;
+	bool sdn_ready : 1;
+	bool initialized : 1;
+	bool wait_initialized : 1;
+};
+
+struct cached_pin {
+	char *id;
+	char *pin;
 };
 
 struct msisdn_set_request {
@@ -151,6 +187,8 @@ static void sim_own_numbers_update(struct ofono_sim *sim);
 
 static GSList *g_drivers = NULL;
 
+static GSList *cached_pins = NULL;
+
 static const char *sim_passwd_name(enum ofono_sim_password_type type)
 {
 	return passwd_name[type];
@@ -324,8 +362,8 @@ static char **get_service_numbers(GSList *service_numbers)
 static void service_number_free(gpointer pointer)
 {
 	struct service_number *num = pointer;
-	g_free(num->id);
-	g_free(num);
+	l_free(num->id);
+	l_free(num);
 }
 
 static void call_state_watches(struct ofono_sim *sim)
@@ -385,6 +423,10 @@ static DBusMessage *sim_get_properties(DBusConnection *conn,
 		ofono_dbus_dict_append(&dict, "ServiceProviderName",
 					DBUS_TYPE_STRING, &sim->spn);
 
+	if (sim->impi)
+		ofono_dbus_dict_append(&dict, "ImsPrivateIdentity",
+					DBUS_TYPE_STRING, &sim->impi);
+
 	fdn = sim->fixed_dialing;
 	ofono_dbus_dict_append(&dict, "FixedDialing", DBUS_TYPE_BOOLEAN, &fdn);
 
@@ -435,6 +477,13 @@ static DBusMessage *sim_get_properties(DBusConnection *conn,
 	get_pin_retries(sim, &pin_retries_dict, &dbus_retries);
 	ofono_dbus_dict_append_dict(&dict, "Retries", DBUS_TYPE_BYTE,
 							&pin_retries_dict);
+
+	ofono_dbus_dict_append(&dict, "CardSlotCount", DBUS_TYPE_UINT32,
+							&sim->card_slot_count);
+
+	ofono_dbus_dict_append(&dict, "ActiveCardSlot", DBUS_TYPE_UINT32,
+							&sim->active_card_slot);
+
 	g_free(pin_retries_dict);
 	g_free(dbus_retries);
 
@@ -444,6 +493,72 @@ done:
 	return reply;
 }
 
+static struct cached_pin *pin_cache_lookup(const char *iccid)
+{
+	struct cached_pin *c;
+	GSList *l;
+
+	if (cached_pins == NULL)
+		return NULL;
+
+	for (l = cached_pins; l; l = l->next) {
+		c = l->data;
+
+		if (g_strcmp0(iccid, c->id) == 0)
+			return c;
+	}
+
+	return NULL;
+}
+
+static void pin_cache_update(const char *iccid, const char *pin)
+{
+	struct cached_pin *pin_cached = pin_cache_lookup(iccid);
+	struct cached_pin *cpins;
+
+	if (pin_cached != NULL) {
+		g_free(pin_cached->pin);
+		pin_cached->pin = g_strdup(pin);
+		return;
+	}
+
+	cpins = g_new0(struct cached_pin, 1);
+
+	cpins->id = g_strdup(iccid);
+	cpins->pin = g_strdup(pin);
+	cached_pins = g_slist_prepend(cached_pins, cpins);
+}
+
+static void pin_cache_remove(const char *iccid)
+{
+	struct cached_pin *pin_cached = pin_cache_lookup(iccid);
+
+	if (pin_cached == NULL)
+		return;
+
+	cached_pins = g_slist_remove(cached_pins, pin_cached);
+
+	g_free(pin_cached->pin);
+	g_free(pin_cached->id);
+	g_free(pin_cached);
+}
+
+static void pin_cache_enter_cb(const struct ofono_error *error, void *data)
+{
+	struct ofono_sim *sim = data;
+
+	if (error->type != OFONO_ERROR_TYPE_NO_ERROR) {
+		pin_cache_remove(sim->iccid);
+
+		__ofono_sim_recheck_pin(sim);
+
+		return;
+	}
+
+	sim->wait_initialized = true;
+	DBG("Waiting for ofono_sim_initialized_notify");
+}
+
 static void sim_pin_retries_query_cb(const struct ofono_error *error,
 					int retries[OFONO_SIM_PASSWORD_INVALID],
 					void *data)
@@ -553,6 +668,36 @@ static gboolean set_own_numbers(struct ofono_sim *sim,
 	return TRUE;
 }
 
+static void sim_set_slot_callback(const struct ofono_error *error, void *data)
+{
+	struct ofono_sim *sim = data;
+	DBusConnection *conn = ofono_dbus_get_connection();
+	const char *path = __ofono_atom_get_path(sim->atom);
+	DBusMessage *reply;
+
+	if (error->type != OFONO_ERROR_TYPE_NO_ERROR) {
+		DBG("Error setting radio access mode");
+
+		sim->pending_active_card_slot = sim->active_card_slot;
+
+		reply = __ofono_error_from_error(error, sim->pending);
+		__ofono_dbus_pending_reply(&sim->pending, reply);
+
+		return;
+	}
+
+	sim->active_card_slot = sim->pending_active_card_slot;
+
+	reply = dbus_message_new_method_return(sim->pending);
+	__ofono_dbus_pending_reply(&sim->pending, reply);
+
+	ofono_dbus_signal_property_changed(conn, path,
+						OFONO_SIM_MANAGER_INTERFACE,
+						"ActiveCardSlot",
+						DBUS_TYPE_UINT32,
+						&sim->active_card_slot);
+}
+
 static DBusMessage *sim_set_property(DBusConnection *conn, DBusMessage *msg,
 					void *data)
 {
@@ -620,6 +765,37 @@ error:
 
 		if (set_ok)
 			return NULL;
+	} else if (!strcmp(name, "ActiveCardSlot")) {
+		dbus_uint32_t value;
+
+		dbus_message_iter_next(&iter);
+
+		if (sim->driver->set_active_card_slot == NULL)
+			return __ofono_error_not_implemented(msg);
+
+		if (dbus_message_iter_get_arg_type(&iter) != DBUS_TYPE_VARIANT)
+			return __ofono_error_invalid_args(msg);
+
+		dbus_message_iter_recurse(&iter, &var);
+
+		if (dbus_message_iter_get_arg_type(&var) != DBUS_TYPE_UINT32)
+			return __ofono_error_invalid_args(msg);
+
+		dbus_message_iter_get_basic(&var, &value);
+
+		if (value <= 0 || value > sim->card_slot_count)
+			return __ofono_error_invalid_args(msg);
+
+		if (sim->active_card_slot == value)
+			return dbus_message_new_method_return(msg);
+
+		sim->pending = dbus_message_ref(msg);
+		sim->pending_active_card_slot = value;
+
+		sim->driver->set_active_card_slot(sim, value - 1,
+							sim_set_slot_callback,
+							sim);
+		return NULL;
 	}
 
 	return __ofono_error_invalid_args(msg);
@@ -652,6 +828,11 @@ static void sim_locked_cb(struct ofono_sim *sim, gboolean locked)
 						OFONO_SIM_MANAGER_INTERFACE,
 						"LockedPins", DBUS_TYPE_STRING,
 						&locked_pins);
+
+	/* Cache pin only for SIM PIN type */
+	if (g_strcmp0(typestr, "pin") == 0)
+		pin_cache_update(sim->iccid, pin);
+
 	g_strfreev(locked_pins);
 
 	sim_pin_retries_check(sim);
@@ -747,6 +928,14 @@ static DBusMessage *sim_unlock_pin(DBusConnection *conn, DBusMessage *msg,
 static void sim_change_pin_cb(const struct ofono_error *error, void *data)
 {
 	struct ofono_sim *sim = data;
+	const char *typestr;
+	const char *old;
+	const char *new;
+
+	dbus_message_get_args(sim->pending, NULL, DBUS_TYPE_STRING, &typestr,
+					DBUS_TYPE_STRING, &old,
+					DBUS_TYPE_STRING, &new,
+					DBUS_TYPE_INVALID);
 
 	if (error->type != OFONO_ERROR_TYPE_NO_ERROR) {
 		__ofono_dbus_pending_reply(&sim->pending,
@@ -757,6 +946,10 @@ static void sim_change_pin_cb(const struct ofono_error *error, void *data)
 		return;
 	}
 
+	/* Cache pin only for SIM PIN type */
+	if (g_strcmp0(typestr, "pin") == 0)
+		pin_cache_update(sim->iccid, new);
+
 	__ofono_dbus_pending_reply(&sim->pending,
 				dbus_message_new_method_return(sim->pending));
 
@@ -808,8 +1001,14 @@ static DBusMessage *sim_change_pin(DBusConnection *conn, DBusMessage *msg,
 static void sim_enter_pin_cb(const struct ofono_error *error, void *data)
 {
 	struct ofono_sim *sim = data;
+	const char *typestr;
+	const char *pin;
 	DBusMessage *reply;
 
+	dbus_message_get_args(sim->pending, NULL, DBUS_TYPE_STRING, &typestr,
+					DBUS_TYPE_STRING, &pin,
+					DBUS_TYPE_INVALID);
+
 	if (error->type != OFONO_ERROR_TYPE_NO_ERROR)
 		reply = __ofono_error_failed(sim->pending);
 	else
@@ -817,6 +1016,22 @@ static void sim_enter_pin_cb(const struct ofono_error *error, void *data)
 
 	__ofono_dbus_pending_reply(&sim->pending, reply);
 
+	/* If PIN entry fails, then recheck the PIN type */
+	if (sim->initialized || error->type != OFONO_ERROR_TYPE_NO_ERROR)
+		goto recheck;
+
+	/* Cache pin only for SIM PIN type */
+	if (g_strcmp0(typestr, "pin") == 0)
+		pin_cache_update(sim->iccid, pin);
+
+	if (sim->pin_type == OFONO_SIM_PASSWORD_SIM_PIN ||
+			sim->pin_type == OFONO_SIM_PASSWORD_SIM_PUK) {
+		sim->wait_initialized = true;
+		DBG("Waiting for ofono_sim_initialized_notify");
+		return;
+	}
+
+recheck:
 	__ofono_sim_recheck_pin(sim);
 }
 
@@ -886,7 +1101,7 @@ static void sim_get_image_cb(struct ofono_sim *sim,
 	if (cache)
 		sim_fs_cache_image(sim->simfs, (const char *) xpm, id);
 
-	g_free(xpm);
+	l_free(xpm);
 }
 
 static void sim_iidf_read_clut_cb(int ok, int length, int record,
@@ -1244,27 +1459,26 @@ static void sim_sdn_read_cb(int ok, int length, int record,
 	if (sim_adn_parse(data, record_length, &ph, &alpha) == FALSE)
 		goto out;
 
-
 	/* Use phone number if Id is unavailable */
 	if (alpha && alpha[0] == '\0') {
-		g_free(alpha);
+		l_free(alpha);
 		alpha = NULL;
 	}
 
 	if (alpha == NULL)
-		alpha = g_strdup(phone_number_to_string(&ph));
+		alpha = l_strdup(phone_number_to_string(&ph));
 
 	if (sim->service_numbers &&
 			g_slist_find_custom(sim->service_numbers,
 				alpha, service_number_compare)) {
 		ofono_error("Duplicate EFsdn entries for `%s'",
 				alpha);
-		g_free(alpha);
+		l_free(alpha);
 
 		goto out;
 	}
 
-	sdn = g_new(struct service_number, 1);
+	sdn = l_new(struct service_number, 1);
 	sdn->id = alpha;
 	memcpy(&sdn->ph, &ph, sizeof(struct ofono_phone_number));
 
@@ -1278,7 +1492,7 @@ check:
 	/* All records retrieved */
 	if (sim->service_numbers) {
 		sim->service_numbers = g_slist_reverse(sim->service_numbers);
-		sim->sdn_ready = TRUE;
+		sim->sdn_ready = true;
 	}
 
 	if (sim->sdn_ready) {
@@ -1434,6 +1648,69 @@ static void sim_set_ready(struct ofono_sim *sim)
 	call_state_watches(sim);
 }
 
+static void impi_read_cb(int ok, int total_length, int record,
+		const unsigned char *data,
+		int record_length, void *userdata)
+{
+	struct ofono_sim *sim = userdata;
+
+	if (!ok) {
+		DBG("error reading IMPI");
+		return;
+	}
+
+	if (data[0] != 0x80) {
+		DBG("invalid TLV tag 0x%02x", data[0]);
+		return;
+	}
+
+	sim->impi = g_strndup((const char *)data + 2, data[1]);
+}
+
+static void discover_apps_cb(const struct ofono_error *error,
+		const unsigned char *dataobj,
+		int len, void *data)
+{
+	GSList *iter;
+	struct ofono_sim *sim = data;
+
+	if (error->type != OFONO_ERROR_TYPE_NO_ERROR)
+		return;
+
+	sim->aid_list = sim_parse_app_template_entries(dataobj, len);
+
+	iter = sim->aid_list;
+
+	while (iter) {
+		struct sim_app_record *app = iter->data;
+		struct ofono_sim_aid_session *s =
+					g_new0(struct ofono_sim_aid_session, 1);
+
+		s->watches = __ofono_watchlist_new(g_free);
+		s->record = app;
+		s->sim = sim;
+		s->session_id = -1;
+		s->state = SESSION_STATE_INACTIVE;
+		sim->aid_sessions = g_slist_prepend(sim->aid_sessions, s);
+
+		if (app->type == SIM_APP_TYPE_ISIM) {
+			/*
+			 * If an ISIM application is found, we should init
+			 * the FS structure so the ISIM EF's can be accessed.
+			 */
+			sim->simfs_isim = sim_fs_new(sim, sim->driver);
+			sim->isim_context = ofono_sim_context_create_isim(
+					sim);
+			/* attempt to get the NAI from EFimpi */
+			ofono_sim_read_bytes(sim->isim_context,
+					SIM_ISIM_EFIMPI_FILEID, 0, 255, NULL,
+					0, impi_read_cb, sim);
+		}
+
+		iter = g_slist_next(iter);
+	}
+}
+
 static void sim_imsi_obtained(struct ofono_sim *sim, const char *imsi)
 {
 	DBusConnection *conn = ofono_dbus_get_connection();
@@ -1554,7 +1831,7 @@ static void sim_fdn_enabled(struct ofono_sim *sim)
 	const char *path = __ofono_atom_get_path(sim->atom);
 	dbus_bool_t val;
 
-	sim->fixed_dialing = TRUE;
+	sim->fixed_dialing = true;
 
 	val = sim->fixed_dialing;
 	ofono_dbus_signal_property_changed(conn, path,
@@ -1591,8 +1868,7 @@ static void sim_efbdn_info_read_cb(int ok, unsigned char file_status,
 		sim_bdn_enabled(sim);
 
 out:
-	if (sim->fixed_dialing != TRUE &&
-			sim->barred_dialing != TRUE)
+	if (!sim->fixed_dialing && !sim->barred_dialing)
 		sim_retrieve_imsi(sim);
 }
 
@@ -1629,8 +1905,7 @@ static void sim_efadn_info_read_cb(int ok, unsigned char file_status,
 
 out:
 	if (check_bdn_status(sim) != TRUE) {
-		if (sim->fixed_dialing != TRUE &&
-				sim->barred_dialing != TRUE)
+		if (!sim->fixed_dialing && !sim->barred_dialing)
 			sim_retrieve_imsi(sim);
 	}
 }
@@ -1714,8 +1989,7 @@ static void sim_efest_read_cb(int ok, int length, int record,
 		sim_bdn_enabled(sim);
 
 out:
-	if (sim->fixed_dialing != TRUE &&
-			sim->barred_dialing != TRUE)
+	if (!sim->fixed_dialing && !sim->barred_dialing)
 		sim_retrieve_imsi(sim);
 }
 
@@ -1844,6 +2118,12 @@ static void sim_initialize_after_pin(struct ofono_sim *sim)
 {
 	sim->context = ofono_sim_context_create(sim);
 
+	/*
+	 * Discover applications on SIM
+	 */
+	if (sim->driver->list_apps)
+		sim->driver->list_apps(sim, discover_apps_cb, sim);
+
 	ofono_sim_read(sim->context, SIM_EFPHASE_FILEID,
 			OFONO_SIM_FILE_STRUCTURE_TRANSPARENT,
 			sim_efphase_read_cb, sim);
@@ -2046,7 +2326,7 @@ skip_efpl:
 	if (!sim->language_prefs_update)
 		__ofono_sim_recheck_pin(sim);
 
-	sim->language_prefs_update = FALSE;
+	sim->language_prefs_update = false;
 }
 
 static void sim_iccid_read_cb(int ok, int length, int record,
@@ -2093,12 +2373,15 @@ static void sim_efli_efpl_changed(int id, void *userdata)
 	if (sim->efli != NULL) /* This shouldn't happen */
 		return;
 
+	if (sim->language_prefs_update)
+		return;
+
 	if (sim->language_prefs) {
 		g_strfreev(sim->language_prefs);
 		sim->language_prefs = NULL;
 	}
 
-	sim->language_prefs_update = TRUE;
+	sim->language_prefs_update = true;
 
 	ofono_sim_read(sim->early_context, SIM_EFLI_FILEID,
 			OFONO_SIM_FILE_STRUCTURE_TRANSPARENT,
@@ -2178,6 +2461,29 @@ struct ofono_sim_context *ofono_sim_context_create(struct ofono_sim *sim)
 	return sim_fs_context_new(sim->simfs);
 }
 
+struct ofono_sim_context *ofono_sim_context_create_isim(
+		struct ofono_sim *sim)
+{
+	GSList *iter = sim->aid_sessions;
+
+	if (sim == NULL || sim->simfs_isim == NULL)
+		return NULL;
+
+	/* Find the AID */
+	while (iter) {
+		struct ofono_sim_aid_session *session = iter->data;
+
+		if (session->record->type == SIM_APP_TYPE_ISIM) {
+			return sim_fs_context_new_with_aid(sim->simfs_isim,
+					session->record->aid);
+		}
+
+		iter = g_slist_next(iter);
+	}
+
+	return NULL;
+}
+
 void ofono_sim_context_free(struct ofono_sim_context *context)
 {
 	return sim_fs_context_free(context);
@@ -2289,6 +2595,15 @@ const unsigned char *ofono_sim_get_cphs_service_table(struct ofono_sim *sim)
 	return sim->cphs_service_table;
 }
 
+ofono_bool_t __ofono_sim_ust_service_available(struct ofono_sim *sim,
+						int ust_service)
+{
+	if (sim->efust)
+		return sim_ust_is_available(sim->efust, sim->efust_length,
+						ust_service);
+	return FALSE;
+}
+
 ofono_bool_t __ofono_sim_service_available(struct ofono_sim *sim,
 						int ust_service,
 						int sst_service)
@@ -2368,15 +2683,24 @@ static void sim_spn_close(struct ofono_sim *sim)
 		sim->cphs_spn_short_watch = 0;
 	}
 
-	sim->flags &= ~SIM_FLAG_READING_SPN;
+	sim->reading_spn = false;
 
-	g_free(sim->spn);
+	l_free(sim->spn);
 	sim->spn = NULL;
 
-	g_free(sim->spn_dc);
+	l_free(sim->spn_dc);
 	sim->spn_dc = NULL;
 }
 
+static void aid_session_free(gpointer data)
+{
+	struct ofono_sim_aid_session *session = data;
+
+	__ofono_watchlist_free(session->watches);
+
+	g_free(session);
+}
+
 static void sim_free_main_state(struct ofono_sim *sim)
 {
 	if (sim->imsi) {
@@ -2395,7 +2719,7 @@ static void sim_free_main_state(struct ofono_sim *sim)
 	if (sim->service_numbers) {
 		g_slist_free_full(sim->service_numbers, service_number_free);
 		sim->service_numbers = NULL;
-		sim->sdn_ready = FALSE;
+		sim->sdn_ready = false;
 	}
 
 	if (sim->efust) {
@@ -2432,8 +2756,8 @@ static void sim_free_main_state(struct ofono_sim *sim)
 	g_free(sim->iidf_image);
 	sim->iidf_image = NULL;
 
-	sim->fixed_dialing = FALSE;
-	sim->barred_dialing = FALSE;
+	sim->fixed_dialing = false;
+	sim->barred_dialing = false;
 
 	sim_spn_close(sim);
 
@@ -2441,6 +2765,28 @@ static void sim_free_main_state(struct ofono_sim *sim)
 		ofono_sim_context_free(sim->context);
 		sim->context = NULL;
 	}
+
+	if (sim->isim_context) {
+		ofono_sim_context_free(sim->isim_context);
+		sim->isim_context = NULL;
+	}
+
+	if (sim->impi)
+		g_free(sim->impi);
+
+	if (sim->aid_list) {
+		g_slist_free_full(sim->aid_list,
+					(GDestroyNotify) sim_app_record_free);
+		sim->aid_list = NULL;
+	}
+
+	if (sim->aid_sessions) {
+		g_slist_free_full(sim->aid_sessions, aid_session_free);
+		sim->aid_sessions = NULL;
+	}
+
+	sim->initialized = false;
+	sim->wait_initialized = false;
 }
 
 static void sim_free_state(struct ofono_sim *sim)
@@ -2577,10 +2923,27 @@ void ofono_sim_inserted_notify(struct ofono_sim *sim, ofono_bool_t inserted)
 		sim->pin_retries[OFONO_SIM_PASSWORD_SIM_PIN2] = -1;
 		sim->pin_retries[OFONO_SIM_PASSWORD_SIM_PUK2] = -1;
 
+		pin_cache_remove(sim->iccid);
+
 		sim_free_state(sim);
 	}
 }
 
+void ofono_sim_initialized_notify(struct ofono_sim *sim)
+{
+	if (sim->state != OFONO_SIM_STATE_INSERTED &&
+				sim->state != OFONO_SIM_STATE_LOCKED_OUT)
+		return;
+
+	sim->initialized = true;
+
+	if (!sim->wait_initialized)
+		return;
+
+	sim->wait_initialized = false;
+	__ofono_sim_recheck_pin(sim);
+}
+
 unsigned int ofono_sim_add_state_watch(struct ofono_sim *sim,
 					ofono_sim_state_event_cb_t notify,
 					void *data, ofono_destroy_func destroy)
@@ -2632,7 +2995,7 @@ static inline void spn_watches_notify(struct ofono_sim *sim)
 	if (sim->spn_watches->items)
 		g_slist_foreach(sim->spn_watches->items, spn_watch_cb, sim);
 
-	sim->flags &= ~SIM_FLAG_READING_SPN;
+	sim->reading_spn = false;
 }
 
 static void sim_spn_set(struct ofono_sim *sim, const void *data, int length,
@@ -2641,10 +3004,10 @@ static void sim_spn_set(struct ofono_sim *sim, const void *data, int length,
 	DBusConnection *conn = ofono_dbus_get_connection();
 	const char *path = __ofono_atom_get_path(sim->atom);
 
-	g_free(sim->spn);
+	l_free(sim->spn);
 	sim->spn = NULL;
 
-	g_free(sim->spn_dc);
+	l_free(sim->spn_dc);
 	sim->spn_dc = NULL;
 
 	if (data == NULL)
@@ -2673,13 +3036,13 @@ static void sim_spn_set(struct ofono_sim *sim, const void *data, int length,
 	}
 
 	if (strlen(sim->spn) == 0) {
-		g_free(sim->spn);
+		l_free(sim->spn);
 		sim->spn = NULL;
 		goto notify;
 	}
 
 	if (dc)
-		sim->spn_dc = g_memdup(dc, 1);
+		sim->spn_dc = l_memdup(dc, 1);
 
 notify:
 	if (sim->spn)
@@ -2748,10 +3111,10 @@ static void sim_spn_changed(int id, void *userdata)
 {
 	struct ofono_sim *sim = userdata;
 
-	if (sim->flags & SIM_FLAG_READING_SPN)
+	if (sim->reading_spn)
 		return;
 
-	sim->flags |= SIM_FLAG_READING_SPN;
+	sim->reading_spn = true;
 	ofono_sim_read(sim->context, SIM_EFSPN_FILEID,
 			OFONO_SIM_FILE_STRUCTURE_TRANSPARENT,
 			sim_spn_read_cb, sim);
@@ -2804,7 +3167,7 @@ ofono_bool_t ofono_sim_add_spn_watch(struct ofono_sim *sim, unsigned int *id,
 		return TRUE;
 	}
 
-	if (sim->flags & SIM_FLAG_READING_SPN)
+	if (sim->reading_spn)
 		return TRUE;
 
 	((ofono_sim_spn_cb_t) item->notify)(sim->spn, sim->spn_dc,
@@ -2835,7 +3198,8 @@ static void sim_pin_query_cb(const struct ofono_error *error,
 	struct ofono_sim *sim = data;
 	DBusConnection *conn = ofono_dbus_get_connection();
 	const char *path = __ofono_atom_get_path(sim->atom);
-	const char *pin_name;
+	struct cached_pin *cpins = pin_cache_lookup(sim->iccid);
+	const char *pin_name = sim_passwd_name(pin_type);
 	char **locked_pins;
 	gboolean lock_changed;
 
@@ -2848,7 +3212,6 @@ static void sim_pin_query_cb(const struct ofono_error *error,
 
 	if (sim->pin_type != pin_type) {
 		sim->pin_type = pin_type;
-		pin_name = sim_passwd_name(pin_type);
 
 		if (pin_type != OFONO_SIM_PASSWORD_NONE &&
 				password_is_pin(pin_type) == FALSE)
@@ -2879,6 +3242,10 @@ static void sim_pin_query_cb(const struct ofono_error *error,
 						&pin_name);
 	}
 
+	if (g_strcmp0(pin_name, "pin") == 0 && cpins != NULL)
+		sim->driver->send_passwd(sim, cpins->pin,
+						pin_cache_enter_cb, sim);
+
 	switch (pin_type) {
 	case OFONO_SIM_PASSWORD_NONE:
 	case OFONO_SIM_PASSWORD_SIM_PIN2:
@@ -2904,7 +3271,6 @@ static void sim_pin_query_cb(const struct ofono_error *error,
 		if (sim->state == OFONO_SIM_STATE_READY)
 			break;
 
-		/* Fall through */
 		sim_initialize_after_pin(sim);
 		break;
 	default:
@@ -2987,6 +3353,8 @@ static void sim_remove(struct ofono_atom *atom)
 
 	sim_fs_free(sim->simfs);
 	sim->simfs = NULL;
+	sim_fs_free(sim->simfs_isim);
+	sim->simfs_isim = NULL;
 
 	g_free(sim);
 }
@@ -3015,6 +3383,9 @@ struct ofono_sim *ofono_sim_create(struct ofono_modem *modem,
 	for (i = 0; i < OFONO_SIM_PASSWORD_INVALID; i++)
 		sim->pin_retries[i] = -1;
 
+	sim->active_card_slot = 1;
+	sim->card_slot_count = 1;
+
 	for (l = g_drivers; l; l = l->next) {
 		const struct ofono_sim_driver *drv = l->data;
 
@@ -3099,8 +3470,6 @@ void ofono_sim_register(struct ofono_sim *sim)
 	sim->spn_watches = __ofono_watchlist_new(g_free);
 	sim->simfs = sim_fs_new(sim, sim->driver);
 
-	__ofono_atom_register(sim->atom, sim_unregister);
-
 	ofono_sim_add_state_watch(sim, sim_ready, sim, NULL);
 
 	if (sim->state > OFONO_SIM_STATE_NOT_PRESENT)
@@ -3109,6 +3478,16 @@ void ofono_sim_register(struct ofono_sim *sim)
 	sim->hfp_watch = __ofono_modem_add_atom_watch(modem,
 					OFONO_ATOM_TYPE_EMULATOR_HFP,
 					emulator_hfp_watch, sim, NULL);
+
+	__ofono_atom_register(sim->atom, sim_unregister);
+}
+
+void __ofono_sim_clear_cached_pins(struct ofono_sim *sim)
+{
+	if (cached_pins == NULL)
+		return;
+
+	pin_cache_remove(sim->iccid);
 }
 
 void ofono_sim_remove(struct ofono_sim *sim)
@@ -3210,17 +3589,17 @@ static void sim_file_changed_flush(struct ofono_sim *sim, int id)
 	sim_fs_cache_flush_file(sim->simfs, id);
 }
 
-void __ofono_sim_refresh(struct ofono_sim *sim, GSList *file_list,
-			ofono_bool_t full_file_change, ofono_bool_t naa_init)
+void __ofono_sim_refresh(struct ofono_sim *sim, struct l_queue *files,
+				bool full_file_change, bool naa_init)
 {
-	GSList *l;
-	gboolean reinit_naa = naa_init || full_file_change;
+	const struct l_queue_entry *l;
+	bool reinit_naa = naa_init || full_file_change;
 
 	/*
 	 * Check if any files used in SIM initialisation procedure
 	 * are affected, except EFiccid, EFpl, EFli.
 	 */
-	for (l = file_list; l; l = l->next) {
+	for (l = l_queue_get_entries(files); l; l = l->next) {
 		struct stk_file *file = l->data;
 		uint32_t mf, df, ef;
 
@@ -3261,7 +3640,7 @@ void __ofono_sim_refresh(struct ofono_sim *sim, GSList *file_list,
 	if (full_file_change)
 		sim_fs_cache_flush(sim->simfs);
 	else {
-		for (l = file_list; l; l = l->next) {
+		for (l = l_queue_get_entries(files); l; l = l->next) {
 			struct stk_file *file = l->data;
 			int id = (file->file[file->len - 2] << 8) |
 				(file->file[file->len - 1] << 0);
@@ -3286,7 +3665,7 @@ void __ofono_sim_refresh(struct ofono_sim *sim, GSList *file_list,
 	if (full_file_change)
 		sim_fs_notify_file_watches(sim->simfs, -1);
 	else {
-		for (l = file_list; l; l = l->next) {
+		for (l = l_queue_get_entries(files); l; l = l->next) {
 			struct stk_file *file = l->data;
 			int id = (file->file[file->len - 2] << 8) |
 				(file->file[file->len - 1] << 0);
@@ -3295,3 +3674,219 @@ void __ofono_sim_refresh(struct ofono_sim *sim, GSList *file_list,
 		}
 	}
 }
+
+const char *__ofono_sim_get_impi(struct ofono_sim *sim)
+{
+	return sim->impi;
+}
+
+static void open_channel_cb(const struct ofono_error *error, int session_id,
+		void *data);
+
+static void close_channel_cb(const struct ofono_error *error, void *data)
+{
+	struct ofono_sim_aid_session *session = data;
+
+	if (error->type != OFONO_ERROR_TYPE_NO_ERROR)
+		DBG("session %d failed to close", session->session_id);
+
+	if (g_slist_length(session->watches->items) > 0 &&
+				session->state == SESSION_STATE_OPENING) {
+		/*
+		 * An atom requested to open during a close, we can re-open
+		 * here.
+		 */
+		session->sim->driver->open_channel(session->sim,
+				session->record->aid, open_channel_cb,
+				session);
+		return;
+	}
+
+	session->state = SESSION_STATE_INACTIVE;
+}
+
+static void open_channel_cb(const struct ofono_error *error, int session_id,
+		void *data)
+{
+	struct ofono_sim_aid_session *session = data;
+	GSList *iter = session->watches->items;
+	ofono_bool_t active = TRUE;
+
+	if (error->type != OFONO_ERROR_TYPE_NO_ERROR) {
+		session->state = SESSION_STATE_INACTIVE;
+		session->session_id = 0;
+		active = FALSE;
+		goto end;
+	}
+
+	if (g_slist_length(iter) == 0) {
+		/*
+		 * All watchers stopped watching before the channel could open.
+		 * Close the channel.
+		 */
+		session->state = SESSION_STATE_CLOSING;
+		session->sim->driver->close_channel(session->sim,
+				session->session_id, close_channel_cb, session);
+		return;
+	}
+
+	session->session_id = session_id;
+	session->state = SESSION_STATE_OPEN;
+end:
+	/*
+	 * Notify any watchers, after this point, all future watchers will be
+	 * immediately notified with the session ID.
+	 */
+	while (iter) {
+		struct ofono_watchlist_item *item = iter->data;
+		ofono_sim_session_event_cb_t notify = item->notify;
+
+		notify(active, session->session_id, item->notify_data);
+
+		iter = g_slist_next(iter);
+	}
+}
+
+unsigned int __ofono_sim_add_session_watch(
+		struct ofono_sim_aid_session *session,
+		ofono_sim_session_event_cb_t notify, void *data,
+		ofono_destroy_func destroy)
+{
+	struct ofono_watchlist_item *item;
+
+	DBG("%p", session);
+
+	if (session == NULL)
+		return 0;
+
+	if (notify == NULL)
+		return 0;
+
+	item = g_new0(struct ofono_watchlist_item, 1);
+
+	item->notify = notify;
+	item->destroy = destroy;
+	item->notify_data = data;
+
+	if (g_slist_length(session->watches->items) == 0 &&
+			session->state == SESSION_STATE_INACTIVE) {
+		/*
+		 * If the session is inactive and there are no watchers, open
+		 * a new session.
+		 */
+		session->state = SESSION_STATE_OPENING;
+		session->sim->driver->open_channel(session->sim,
+				session->record->aid, open_channel_cb,
+				session);
+	} else if (session->state == SESSION_STATE_OPEN) {
+		/*
+		 * Session is already open and available, just call the
+		 * notify callback immediately.
+		 */
+		notify(TRUE, session->session_id, data);
+	} else if (session->state == SESSION_STATE_CLOSING) {
+		/*
+		 * There is a pending close, the close callback will handle
+		 * re-opening the session.
+		 */
+		session->state = SESSION_STATE_OPENING;
+	}
+
+	return __ofono_watchlist_add_item(session->watches, item);
+}
+
+void __ofono_sim_remove_session_watch(struct ofono_sim_aid_session *session,
+		unsigned int id)
+{
+	__ofono_watchlist_remove_item(session->watches, id);
+
+	if (g_slist_length(session->watches->items) == 0) {
+		/* last watcher, close session */
+		session->state = SESSION_STATE_CLOSING;
+		session->sim->driver->close_channel(session->sim,
+				session->session_id, close_channel_cb, session);
+	}
+}
+
+struct ofono_sim_aid_session *__ofono_sim_get_session_by_aid(
+		struct ofono_sim *sim, unsigned char *aid)
+{
+	GSList *iter = sim->aid_sessions;
+
+	while (iter) {
+		struct ofono_sim_aid_session *session = iter->data;
+
+		if (!memcmp(session->record->aid, aid, 16))
+			return session;
+
+		iter = g_slist_next(iter);
+	}
+
+	return NULL;
+}
+
+struct ofono_sim_aid_session *__ofono_sim_get_session_by_type(
+		struct ofono_sim *sim, enum sim_app_type type)
+{
+	GSList *iter = sim->aid_sessions;
+
+	while (iter) {
+		struct ofono_sim_aid_session *session = iter->data;
+
+		if (session->record->type == type)
+			return session;
+
+		iter = g_slist_next(iter);
+	}
+
+	return NULL;
+}
+
+int __ofono_sim_session_get_id(struct ofono_sim_aid_session *session)
+{
+	return session->session_id;
+}
+
+enum sim_app_type __ofono_sim_session_get_type(
+		struct ofono_sim_aid_session *session)
+{
+	return session->record->type;
+}
+
+unsigned char *__ofono_sim_session_get_aid(
+		struct ofono_sim_aid_session *session)
+{
+	return session->record->aid;
+}
+
+GSList *__ofono_sim_get_aid_list(struct ofono_sim *sim)
+{
+	return sim->aid_list;
+}
+
+int ofono_sim_logical_access(struct ofono_sim *sim, int session_id,
+		unsigned char *pdu, unsigned int len,
+		ofono_sim_logical_access_cb_t cb, void *data)
+{
+	if (!sim->driver->logical_access)
+		return -ENOTSUP;
+
+	if (session_id <= 0)
+		return -EINVAL;
+
+	sim->driver->logical_access(sim, session_id, pdu, len, cb, data);
+
+	return 0;
+}
+
+void ofono_sim_set_card_slot_count(struct ofono_sim *sim, unsigned int val)
+{
+	if (sim)
+		sim->card_slot_count = val;
+}
+
+void ofono_sim_set_active_card_slot(struct ofono_sim *sim, unsigned int val)
+{
+	if (sim)
+		sim->active_card_slot = val;
+}
